/*  Jeremy Smith
 *  G00334574
 *  CS 688 Project Fall 2013
 *  Submitted: 12/9/2013
 */

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import javax.imageio.ImageIO;

public class CS688_Project 
{
    private final int INPUT_NODES = 4;      // number of input nodes
    private final int HIDDEN_NODES = 4;     // number of hidden nodes
    private final int OUTPUT_NODES = 4;     // number of output nodes
    private final float ALPHA = 0.01f;      // learning rate
    private final float BETA = 0.5f;        // momentum
    private final int ATTEMPTS = 10;        // number of training attempts
    private float [] x;                     // input node values
    private float [] h;                     // hidden node values
    private float [] o;                     // output node values
    private float [] t;                     // target values
    private float [] dk;                    // error on output nodes
    private float [] dj;                    // error on hidden nodes
    private float [][] inputValues;         // set of input values to the NN
    private float [][] targetValues;        // set of target values for the NN
    private float [][] wh;                  // weights from input to hidden layer
    private float [][] wo;                  // weights from hidden to output layer
    private float [][] delta_wh;            // change for each hidden weight
    private float [][] delta_wo;            // change for each output weight
    private float prevError;                // total error for previous training epoc
    private float totalError;               // total error for current training epoc
    
    public static void main(String[] args) 
    {
        CS688_Project network = new CS688_Project();
        
        // load training data from files
        //if(network.loadTargetData("C:\\CS 688\\Pictures\\Training\\targets.txt", 80)) { System.exit(1); }
        //if(network.loadImageData("C:\\CS 688\\Pictures\\Training\\training ", 80)) { System.exit(1); }
        if(network.loadRawData("C:\\CS 688\\training.txt", 80)) { System.exit(1); }
        network.trainNetwork();
        
        // load testing data from files
        //if(network.loadTargetData("C:\\CS 688\\Pictures\\Testing\\targets.txt", 300)) { System.exit(1); }
        //if(network.loadImageData("C:\\CS 688\\Pictures\\Testing\\testing ", 300)) { System.exit(1); }
        if(network.loadRawData("C:\\CS 688\\testing.txt", 300)) { System.exit(1); }
        network.testNetwork();
    }
    
    public boolean loadImageData(String filename, int inputs)
    {
        int redX, redY, greenX, greenY, blueX, blueY, redCount, greenCount, blueCount, width, height;
        BufferedImage image = null;
        inputValues = new float[inputs][INPUT_NODES];
        
        for(int i=0; i<inputs; i++)
        {
            try 
            { 
                image = ImageIO.read(new File(filename + "(" + i + ").png")); 
            } 
            catch (IOException e) 
            { 
                return true; 
            }

            width = image.getWidth();
            height = image.getHeight();
            redX = 0;
            redY = 0;
            redCount = 0;
            greenX = 0;
            greenY = 0;
            greenCount = 0;
            blueX = 0;
            blueY = 0;
            blueCount = 0;

            // Loop through each pixel in the image
            for(int x = 0; x < width; x++)
            {
                for(int y = 0; y < height; y++)
                {
                    // get the pixel's color
                    int color = image.getRGB(x, y);
                    int red = (color & 0x00ff0000) >> 16;
                    int green = (color & 0x0000ff00) >> 8;
                    int blue = color & 0x000000ff;

                    // determine if the pixel corresponds to an expected color and get its coordinates
                    if(red > 80 && green < 10 && blue < 10)
                    {
                        redX += x;
                        redY += y;
                        redCount++;
                    }
                    if(red < 20 && green > 60 && blue < 20)
                    {
                        greenX += x;
                        greenY += y;
                        greenCount++;
                    }
                    if(red < 10 && green < 10 && blue > 40)
                    {
                        blueX += x;
                        blueY += y;
                        blueCount++;
                    }
                }
            }

            if(width == 0 || height == 0 || redCount == 0 || greenCount == 0 || blueCount == 0)
            {
                System.out.println("Error on file " + i);
                return true;
            }

            // average coordinates of all pixels of each color
            float redXAvg = ((float)redX / (float)redCount);
            float redYAvg = ((float)redY / (float)redCount);
            float greenXAvg = ((float)greenX / (float)greenCount);
            float greenYAvg = ((float)greenY / (float)greenCount);
            float blueXAvg = ((float)blueX / (float)blueCount);
            float blueYAvg = ((float)blueY / (float)blueCount);

            // use the averages of red and blue with respect to green to get the inputs
            inputValues[i][0] = redXAvg - greenXAvg;
            inputValues[i][1] = redYAvg - greenYAvg;
            inputValues[i][2] = blueXAvg - greenXAvg;
            inputValues[i][3] = blueYAvg - greenYAvg;
        }
        
        return false;
    }
    
    public boolean loadTargetData(String filename, int inputs)
    {
        FileReader reader = null;
        BufferedReader buff = null;
        targetValues = new float[inputs][OUTPUT_NODES];
        
        try 
        { 
            reader = new FileReader(new File(filename)); 
        } 
        catch (FileNotFoundException ex) 
        { 
            return true; 
        }
        buff = new BufferedReader(reader);
        
        for(int i = 0; i < inputs; i++)
        {
            String lineRead = "";
            try { lineRead = buff.readLine(); } 
            catch (IOException ex) { return true; }
            
            for(int k = 0; k < OUTPUT_NODES; k++)
            {
                targetValues[i][k] = Float.parseFloat(lineRead.split(",")[k+1]);
            }
        }
        
        return false;
    }
    
    public boolean loadRawData(String filename, int inputs)
    {
        FileReader reader = null;
        BufferedReader buff = null;
        inputValues = new float[inputs][INPUT_NODES];
        targetValues = new float[inputs][OUTPUT_NODES];
        
        try 
        { 
            reader = new FileReader(new File(filename)); 
        } 
        catch (FileNotFoundException ex) 
        { 
            return true; 
        }
        buff = new BufferedReader(reader);
        
        for(int i = 0; i < inputs; i++)
        {
            String lineRead = "";
            try { lineRead = buff.readLine(); } 
            catch (IOException ex) { return true; }
            
            for(int k = 0; k < INPUT_NODES; k++)
            {
                inputValues[i][k] = Float.parseFloat(lineRead.split(" ")[k]);
            }
            for(int k = 0; k < OUTPUT_NODES; k++)
            {
                targetValues[i][k] = Float.parseFloat(lineRead.split(" ")[k+INPUT_NODES]);
            }
        }
        
        return false;
    }
    
    public void trainNetwork()
    {
        float [][] wo_final = new float[HIDDEN_NODES+1][OUTPUT_NODES];
        float [][] wh_final = new float[INPUT_NODES+1][HIDDEN_NODES];
        float minError = 100.0f;
        
        for(int a = 0; a < ATTEMPTS; a++)
        {
            int count = 0;
            totalError = 1.0f;
            prevError = 0.0f;

            // initialize the network nodes and randomize the network weights
            initialize();

            while(totalError != prevError)
            {
                prevError = totalError;
                totalError = 0;
                // Perform training
                for(int in = 0; in < inputValues.length; in++)
                {
                    // read inputs and targets
                    readInputAndTargets(in);

                    // forward propagation
                    for(int j = 0; j < HIDDEN_NODES; j++)
                    {
                        h[j] = calculateHiddenNode(j);
                    }
                    for(int k = 0; k < OUTPUT_NODES; k++)
                    {
                        o[k] = calculateOutputNode(k);
                    }
                    // compute error
                    float error = 0.0f;
                    for(int k = 0; k < OUTPUT_NODES; k++)
                    {
                        error += (t[k] - o[k]) * (t[k] - o[k]);
                    }
                    totalError += error / OUTPUT_NODES;

                    // back propagation
                    for(int k = 0; k < OUTPUT_NODES; k++)
                    {   
                        dk[k] = calculateOutputError(k);

                        for(int j = 0; j <= HIDDEN_NODES; j++)
                        {
                            backpropagateOutputWeight(k, j);
                        }
                    }
                    for(int j = 0; j < HIDDEN_NODES; j++)
                    {
                        dj[j] = calculateHiddenError(j);

                        for(int i = 0; i <= INPUT_NODES; i++)
                        {
                            backpropagateHiddenWeight(j, i);
                        }
                    }
                }
                count++;
            }

            // Save weights if they produce the minimum error in the ensemble
            if(totalError < minError)
            {
                minError = totalError;
                for(int i=0; i<INPUT_NODES+1; i++)
                {
                    for(int j = 0; j< HIDDEN_NODES; j++)
                    {
                        wh_final[i][j] = wh[i][j];
                    }
                }
                for(int j=0; j<HIDDEN_NODES+1; j++)
                {
                    for(int k = 0; k< OUTPUT_NODES; k++)
                    {
                        wo_final[j][k] = wo[j][k];
                    }
                }
            }
        }
        
        System.out.printf("Training Error: %f\n\n", minError);
        // Apply weights with the minimum error in the ensemble
        for(int i=0; i<INPUT_NODES+1; i++)
        {
            System.out.print("wh" + i + "\t");
            for(int j = 0; j< HIDDEN_NODES; j++)
            {
                wh[i][j] = wh_final[i][j];
                System.out.printf("%f\t",wh[i][j]);
            }
            System.out.print("\n");
        }
        System.out.print("\n");
        for(int j=0; j<HIDDEN_NODES+1; j++)
        {
            System.out.print("wo" + j + "\t");
            for(int k = 0; k< OUTPUT_NODES; k++)
            {
                wo[j][k] = wo_final[j][k];
                System.out.printf("%f\t",wo[j][k]);
            }
            System.out.print("\n");
        }
    }
    
    public void testNetwork()
    {        
        totalError = 0;
        
        // Perform training
        for(int in = 0; in < inputValues.length; in++)
        {
            // read inputs and targets
            readInputAndTargets(in);

            // forward propagation
            for(int j = 0; j < HIDDEN_NODES; j++)
            {
                h[j] = calculateHiddenNode(j);
            }
            
            for(int k = 0; k < OUTPUT_NODES; k++)
            {
                o[k] = calculateOutputNode(k);
            }
            
            // compute error
            float error = 0.0f;
            for(int k = 0; k < OUTPUT_NODES; k++)
            {
                error += (t[k] - o[k]) * (t[k] - o[k]);
            }
            totalError += error / OUTPUT_NODES;
        }
        
        System.out.print("\n");
        System.out.printf("Testing Error: %f\n", totalError);
    }
    
    private void readInputAndTargets(int in)
    {
        for(int i = 0; i < INPUT_NODES; i++)
        {
            x[i] = inputValues[in][i];
        }
        for(int k = 0; k < OUTPUT_NODES; k++)
        {
            t[k] = targetValues[in][k];
        }
        
        x[INPUT_NODES] = 1.0f;
            h[HIDDEN_NODES] = 1.0f;
    }
    
    private void initialize()
    {
        x = new float[INPUT_NODES+1];
        h = new float[HIDDEN_NODES+1];
        o = new float[OUTPUT_NODES];
        t = new float[OUTPUT_NODES];
        dk = new float[OUTPUT_NODES];
        dj = new float[HIDDEN_NODES];
        wh = new float[INPUT_NODES+1][HIDDEN_NODES];
        wo = new float[HIDDEN_NODES+1][OUTPUT_NODES];
        delta_wh = new float[INPUT_NODES+1][HIDDEN_NODES];
        delta_wo = new float[HIDDEN_NODES+1][OUTPUT_NODES];
        
        randomizeWeights();
    }
    
    private void randomizeWeights()
    {
        for(int i=0; i< wh.length; i++)
        {
            for(int j=0; j< wh[0].length; j++)
            {
                wh[i][j] = (float)Math.random();
                delta_wh[i][j] = 0.0f;
            }
        }
        
        for(int j=0; j< wo.length; j++)
        {
            for(int k=0; k< wo[0].length; k++)
            {
                wo[j][k] = (float)Math.random();
                delta_wo[j][k] = 0.0f;
            }
        }
    }
    
    private float calculateHiddenNode(int j)
    {
        float out, z, e;
        
        z = wh[INPUT_NODES][j];
        
        for(int i = 0; i < INPUT_NODES; i++)
        {
            z += x[i] * wh[i][j];
        }
        
        e = (float)Math.exp((double)-z);
        
        out = 1.0f / (1.0f + e);
        
        return out;
    }
    
    private float calculateOutputNode(int k)
    {
        float out, z, e;
        
        z = wo[HIDDEN_NODES][k];
        
        for(int j = 0; j < HIDDEN_NODES; j++)
        {
            z += h[j] * wo[j][k];
        }
        
        e = (float)Math.exp((double)-z);
        
        out = 1.0f / (1.0f + e);
        
        return out;
    }
    
    private float calculateOutputError(int k) 
    {
        return o[k] * (1 - o[k]) * (o[k] - t[k]);
    }
    
    private void backpropagateOutputWeight(int k, int j)
    {
        delta_wo[j][k] = -ALPHA * dk[k] * h[j] + BETA * delta_wo[j][k];
        
        wo[j][k] = wo[j][k] + delta_wo[j][k];       
    }
    
    private float calculateHiddenError(int j)
    {
        float out;
        
        out = dk[0] * wo[j][0];
        
        for(int k = 1; k < OUTPUT_NODES; k++)
        {
            out += dk[k] * wo[j][k];
        }
        
        out *= h[j] * (1 - h[j]);
        
        return out;
    }
    
    private void backpropagateHiddenWeight(int j, int i)
    {
        delta_wh[i][j] = -ALPHA * dj[j] * x[i] + BETA * delta_wh[i][j];
        
        wh[i][j] = wh[i][j] + delta_wh[i][j];
    }
}
